require("util.extension.table")
local json = require("util.common.JSON")
local global_save_file = "config/settings.json"
local lfs = import("lfs")
local __settings = {}
return function(save_file)
    save_file = save_file or global_save_file
    if __settings[save_file] then
        return __settings[save_file]
    end
    local settings = {}
    local loaded = false
    local function load()
        local f = _G.io.open((save_file or global_save_file), "r")
        if f then
            local content = f:read("*a")
            f:close()
            if not pcall(function()
                local t = json.decode(content)
                if t then
                    settings = t
                end
            end) then
                settings = {}
            end
        end
    end

    local function save()
        if table.empty(settings) then
            return
        end
        if (save_file or global_save_file):find("(.+)[/\\].+") then
            lfs.mkdir((save_file or global_save_file):match("(.+)[/\\].+"))
        end
        local f = _G.io.open((save_file or global_save_file), "w")
        if f then
            f:write(json.encode(settings))
            f:close()
        end
    end

    local function set_setting(name, value)
        if not loaded then load() loaded = true end
        if settings[name] == value then
            return value
        end
        settings[name] = value
        save()
        return value
    end

    local function get_setting(name, default)
        if not loaded then load() loaded = true end
        if settings[name] == nil and default ~= nil then
            return default
        end
        return settings[name]
    end

    __settings[save_file] = {
        set = set_setting,
        get = get_setting,
        base = settings,
        save = save,
        set_file = function(file)
            save_file = file
            load()
        end,
    }

    return __settings[save_file]
end