local watcher = {
    watched = {},
    callbacks = {}
}

local lfs = import("lfs")
local events = require("util.common.events")

function watcher.watch(path, cb)
    local attr = lfs.attributes(path)
    if attr and attr.mode == "directory" then
        for file in lfs.dir(path) do
            if file ~= "." and file ~= ".." and file ~= "modules" and file ~= "assets" and file:find("%.lua$") then
                watcher.watch(path .. "/" .. file, cb)
            end
        end
    else
        watcher.watched[#watcher.watched + 1] = {
            path = path,
            last = lfs.attributes(path).modification,
            callback = cb
        }
        print("watching " .. path)
    end
end

function watcher.unwatch(path)
    for i = 1, #watcher.watched do
        if watcher.watched[i] == path then
            table.remove(watcher.watched, i)
            break
        end
    end
end

function watcher.on_change(cb)
    watcher.callbacks[#watcher.callbacks + 1] = cb
end

events.update += function()
    for i = 1, #watcher.watched do
        local w = watcher.watched[i]
        local attr = lfs.attributes(w.path)
        if attr.modification ~= w.last then
            print("change detected in " .. w.path)
            w.last = attr.modification
            if w.callback then
                w.callback(w.path)
            end
            for j = 1, #watcher.callbacks do
                watcher.callbacks[j](w.path)
            end
        end
    end
end

return watcher