_G.warn = function(s)
    printft(string.format("[Warning] %s", s))
end

---prints formatted
_G.warnf = function(...)
    warn(string.format(...))
end

local function get_timezone_offset(ts)
    local utcdate   = os.date("!*t", ts)
    local localdate = os.date("*t", ts)
    localdate.isdst = false -- this is the trick
    return os.difftime(os.time(localdate), os.time(utcdate))
end

---prints formatted
_G.printft = function (fmt, ...)
    local t = clock() + get_timezone_offset(clock())
    local tstr = string.format("%02d:%02d:%02d.%03d ", (t / 3600) % 24, t / 60 % 60, t % 60, t * 1000 % 1000)
    print(tstr .. string.format(fmt, ...))
end

---prints as hex
_G.printh = function(...)
    local out = ""
    local args = {...}
    local argn = 0
    for i = 1, select("#", ...) do
        if type(args[i]) == "number" then
            out = out .. (out == "" and "%X" or "%10X")
        else
            out = out .. "%10s"
            args[i] = tostring(args[i])
        end
    end
    printft(out, table.unpack(args))
end

---prints as hex
_G.printha = function(str)
    local out = ""
    for i = 1, #str do
        out = out .. string.format("%02X ", str:byte(i))
    end
    printft(out)
end

---prints as hex
_G.printhaft = function(fmt, str)
    local out = ""
    for i = 1, #str do
        out = out .. string.format("%02X ", str:byte(i))
    end
    printft(fmt, out)
end

_G.range = function(start, stop, step)
    if not stop then
        stop = start
        start = 1
    end
    step = step or 1
    return coroutine.wrap(function()
        coroutine.yield(start)
        while start < stop do
            start = start + step
            coroutine.yield(start)
        end
    end)
end

_G.using = function(t)
    if type(t) == "string" then
        t = import(t)
    end
    local env = getfenv(2)
    for k, v in pairs(t) do
        env[k] = v
    end
end