local debug_getupvalue = debug.getupvalue
local debug_getinfo = debug.getinfo
local debug_functrace = function(info)
    local trace = ""
    if type(info) == "function" then
        info = debug_getinfo(info)
    end
    if info.source ~= "=?" then
        if info.linedefined and info.linedefined > 0 then
            trace = trace .. info.source .. ":" .. info.linedefined .. ":"
        else
            trace = trace .. info.source .. ":"
        end
    elseif info.nups > 0 then
        for i = 1, info.nups do
            local _, uv = debug_getupvalue(info.func, i)
            if type(uv) == "table" then
                if uv.chunk_name then
                    if info.linedefined and info.linedefined > 0 then
                        trace = trace .. tostring(uv.chunk_name) .. ":" .. info.linedefined .. ":"
                    else
                        trace = trace .. tostring(uv.chunk_name) .. ":"
                    end
                end
            end
        end
    else
        trace = trace .."?:"
    end
    trace = trace .. " in "
    if info.namewhat and #info.namewhat > 0 then
        trace = trace .. info.namewhat .. " "
    end
    if info.name and #info.name > 0 and info.name[1] ~= '?' then
        trace = trace .. "'" .. info.name .. "' "
    elseif info.what == "main" then
        if info.nups > 0 then
            for i = 1, info.nups do
                local _, uv = debug_getupvalue(info.func, i)
                if type(uv) == "table" then
                    if uv.header then
                        trace = trace
                            .. "module '"
                            .. tostring(uv.header.name)
                            .. "' with id '"
                            .. tostring(uv.header.id) .. "'"
                    end
                end
            end
        else
            trace = trace .. "main chunk"
        end
    else
        trace = trace
            .. (info.namewhat and #info.namewhat > 0 and "" or "function ")
            .. "(defined in lines " .. tostring(info.linedefined) .. "-" .. tostring(info.lastlinedefined) .. ")"
    end
    return trace
end
local debug_traceback = debug.traceback;
local function new_debug_traceback(add_msg, level)
    if debug.current_invoke then
        local prefix = "[" .. tostring(debug.current_invoke) .. "]"
        if add_msg then
            add_msg = prefix .. " " .. add_msg
        else
            add_msg = prefix
        end
    end
    return debug_traceback(add_msg or "", (level or 1) + 1)
end
_G.debug.functrace = debug_functrace
_G.debug.traceback = new_debug_traceback