table.getn = table.getn or function(t)
    return #t
end
table.isarray = table.isarray or function(t)
    local i = 0
    for _ in pairs(t) do
        i = i + 1
        if t[i] == nil then
            return false
        end
    end
    return i == #t
end
table.unpack = table.unpack or unpack
table.is_vector = function(t)
    if #t > 1 and #t <= 4 then
        for i = 1, #t do
            if type(t[i]) ~= "number" then
                return false
            end
        end
        return true
    end
    local indices = {}
    for k, _ in pairs(t) do
        if k == "x" or k == "y" or k == "z" or k == "w" then
            continue
        end
        indices[#indices + 1] = k
    end
    return #indices == 0
end
table.serialize = table.serialize or function(t, trim, indent, cb)
    indent = indent or ""
    local str = ""
    cb = cb or function(append)
        str = str .. append
    end
    if table.is_vector(t) then
        cb(indent .. "(")
        local space = trim and "" or " "
        if #t > 1 then
            for i = 1, #t - 1 do
                cb(t[i] .. "," .. space)
            end
            cb(t[#t])
        else
            cb(t.x .. "," .. space .. t.y)
            if t.z then
                cb("," .. space .. t.z)
            end
            if t.w then
                cb("," .. space .. t.w)
            end
        end
        cb(")")
    else
        local k, v = next(t)
        while k do
            local kt = type(k)
            local vt = type(v)
            if kt == "string" then
                cb(indent .. (tonumber(k[1]) and "[\"" .. k .. "\"]" or k) .. (trim and "=" or " = "))
            elseif kt == "number" then
                cb(indent .. "[" .. k .. "]" .. (trim and "=" or " = "))
            else
                error("Cannot serialize table with key of type " .. kt)
            end
            if vt == "string" then
                cb("\"" .. v .. "\"")
            elseif vt == "number" then
                cb(v)
            elseif vt == "boolean" then
                cb((v and "true" or "false"))
            elseif vt == "table" then
                if table.is_vector(v) then
                    table.serialize(v, trim, indent, cb)
                else
                    cb((trim and "{" or "{\n"))
                    table.serialize(v, trim, indent .. (trim and "" or "  "), cb)
                    cb(indent .. (trim and "}" or "\n}"))
                end
            else
                cb(indent .. tostring(v))
            end
            k, v = next(t, k)
            if k then
                cb(indent .. (trim and "," or ",\n"))
            elseif not trim then
                cb("\n")
            end
        end
    end
    return str
end

local resolve__call = function(t, ...)
    if type(t) == "function" then
        return t(...)
    end
    local mt = getmetatable(t)
    if mt and mt.__call then
        return mt.__call(t, ...)
    end
    if not t[1] then
        printft(debug.traceback())
    end
    return t[1](t[2], ...)
end
local resolve__pcall = function(t, ...)
    if type(t) == "function" then
        return pcall(t, ...)
    end
    local mt = getmetatable(t)
    if mt and mt.__call then
        return pcall(mt.__call, t, ...)
    end
    return pcall(t[1], t[2], ...)
end

table.tostring = function(b)
    return "{" .. (table.fold(b, "(x, y) => (y and tostring(y) .. ', ' or '') .. tostring(x)") or '') .. "}"
end
table.skip = function(b, where, amount)
    if type(where) == "number" then where, amount = "begin", where end
    local newref = {}
    for i = 1 + (where == "begin" and amount or 0), #b - (where == "end" and amount or 0) do
        newref[#newref + 1] = b[i]
    end
    return newref
end
table.invoke = function(b, ...)
    for i = 1, #b do
        resolve__call(b[i], ...)
    end
end
table.pinvoke = function(b, ...)
    for i = 1, #b do
        local status, err = resolve__pcall(b[i], ...)
        if not status then
            printft(err .. "\n" .. debug.functrace(b[i]))
        end
    end
end
table.empty = function(b)
    return next(b) == nil
end
table.count = function(b)
    local c = 0
    for _, _ in pairs(b) do
        c = c + 1
    end
    return c
end
table.at = function(b, idx)
    return b[idx]
end
table.iter = function(b, begin, finish)
    local __idx__ = begin or 0
    return function()
        if __idx__ < finish or #b then
            __idx__ += 1
            return b[__idx__]
        end
    end
end
table.call = function(b, ...)
    if b.id == 1 then
        for i = 1, #b do
            if not resolve__call(b[i], ...) then
                return false
            end
        end
        return true
    elseif b.id == 2 then
        for i = 1, #b do
            if resolve__call(b[i], ...) then
                return true
            end
        end
        return false
    end
    for i = 1, #b do
        local result = resolve__call(b[i], ...)
        if result ~= nil then
            return result
        end
    end
end
table.invokeUntil = function(b, ...)
    for i = 1, #b do
        if resolve__call(b[i], ...) then
            return true
        end
    end
    return false
end
table.invokeUnless = function(b, ...)
    for i = 1, #b do
        if resolve__call(b[i], ...) == false then
            return false
        end
    end
    return true
end
table.add = function(b, o)
    if type(o) == "table" then
        if o.type%"Box" then
            for e in o.Iter() do
                b[#b+1] = e
            end
        elseif #o > 0 then
            for i = 1, #o do
                b[#b+1] = o[i]
            end
        else
            for _, element in pairs(o) do
                b[#b+1] = element
            end
        end
    else
        b[#b+1] = o
    end
    return b
end
table.merge = function(b, o)
    local newref = { }
    for i = 1, #b do
        newref[#newref + 1] = b[i]
    end
    if type(o) == "table" then
        if o.type%"Box" then
            for e in o.Iter() do
                newref[#newref + 1] = e
            end
        elseif #o > 0 then
            for i = 1, #o do
                newref[#newref + 1] = o[i]
            end
        else
            for _, element in pairs(o) do
                newref[#newref + 1] = element
            end
        end
    else
        newref[#newref + 1] = o
    end
    return newref
end
table.copy = function(b)
    local newref = {}
    for i, v in pairs(b) do
        newref[i] = v
    end
    return newref
end
table.deep_copy = function(b)
    local newref = {}
    for i, v in pairs(b) do
        if type(v) == "table" then
            newref[i] = table.deep_copy(v)
        else
            newref[i] = v
        end
    end
    return newref
end
table.erase = function(b, o)
    local newref = {}
    for i = 1, #b do
        if b[i] ~= o then
            newref[#newref + 1] = b[i]
        end
    end
    return newref
end
table.any = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            if t(b[i], ...) then
                return true
            end
        end
    else
        return #b > 0
    end
    return false
end
table.where = function(b, t, ...)
    local newref = {}
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            if t(b[i], ...) then
                newref[#newref + 1] = b[i]
            end
        end
    end
    return newref
end
table.except = function(b, t, ...)
    local newref = { }
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            if not t(b[i], ...) then
                newref[#newref + 1] = b[i]
            end
        end
    end
    return newref
end
table.forEach = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            t(b[i], ...)
        end
    end
    return b
end
table.forEachUntil = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            if t(b[i], ...) then
                return b
            end
        end
    end
    return b
end
table.first = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        for i = 1, #b do
            if t(b[i], ...) then
                return b[i]
            end
        end
    else
        return b[1]
    end
    return nil
end
table.last = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        for i = #b, 1, -1 do
            if t(b[i], ...) then
                return b[i]
            end
        end
    else
        return b[#b]
    end
    return nil
end
table.fold = function(b, t, ...)
    if #b == 0 then
        return nil
    end
    local t_type = type(t)
    if t_type == "function" then
        local o = t(b[1], ...)
        for i = 2, #b do
            o = t(b[i], o, ...)
        end
        return o
    end
    return nil
end
table.clear = function(b)
    for i, v in pairs(b) do
        b[i] = nil
    end
    return b
end
table.apply = function(b, t, ...)
    local newref = {}
    local t_type = type(t)
    if t_type == "table" then
        for _, v in pairs(t) do
            for i = 1, #b do
                newref[i] = v(b[i], ...)
            end
        end
    elseif t_type == "function" then
        for i = 1, #b do
            newref[i] = t(b[i], ...)
        end
    end
    return newref
end
table.order = function(b, t, ...)
    local t_type = type(t)
    if t_type == "function" then
        local sorted = #b == 0
        while not sorted do
            sorted = true
            for i = 1, #b - 1 do
                if t(b[i], b[i + 1], ...) then
                    b[i], b[i + 1] = b[i + 1], b[i]
                    sorted = false
                end
            end
        end
    else
        local sorted = #b == 0
        while not sorted do
            sorted = true
            for i = 1, #b - 1 do
                if b[i] > b[i + 1] then
                    b[i], b[i + 1] = b[i + 1], b[i]
                    sorted = false
                end
            end
        end
    end
    return b
end
table.contains = function(b, o)
    for i = 1, #b do
        if b[i] == o then
            return true
        end
    end
    return false
end